"""
Signal generation stack.

Contrary to most quanutm simulators, C^3 includes a detailed simulation of the control
stack. Each component in the stack and its functions are simulated individually and
combined here.

Example: A local oscillator and arbitrary waveform generator signal
are put through via a mixer device to produce an effective modulated signal.
"""

import copy
from typing import List
import hjson
import numpy as np
import tensorflow as tf
from c3.c3objs import hjson_decode, hjson_encode
from c3.signal.gates import Instruction
from c3.generator.devices import devices as dev_lib


class Generator:
    """
    Generator, creates signal from digital to what arrives to the chip.

    Parameters
    ----------
    devices : list
        Physical or abstract devices in the signal processing chain.
    resolution : np.float64
        Resolution at which continuous functions are sampled.

    """

    def __init__(
        self, devices: dict = None, chains: dict = None, resolution: np.float64 = 0.0
    ):
        self.devices = {}
        if devices:
            self.devices = devices
        self.chains = {}
        if chains:
            self.chains = chains
            self.__check_signal_chains()
        self.resolution = resolution
        self.gen_stacked_signals: dict = None

    def __check_signal_chains(self) -> None:
        for channel, chain in self.chains.items():
            signals = 0
            for device_id in chain:
                signals -= self.devices[device_id].inputs
                signals += self.devices[device_id].outputs
            if signals != 1:
                raise Exception(
                    "C3:ERROR: Signal chain for channel '"
                    + channel
                    + "' contains unmatched number of inputs and outputs."
                )

    def read_config(self, filepath: str) -> None:
        """
        Load a file and parse it to create a Generator object.

        Parameters
        ----------
        filepath : str
            Location of the configuration file

        """
        with open(filepath, "r") as cfg_file:
            cfg = hjson.loads(cfg_file.read(), object_pairs_hook=hjson_decode)
        self.fromdict(cfg)

    def fromdict(self, cfg: dict) -> None:
        for name, props in cfg["Devices"].items():
            props["name"] = name
            dev_type = props.pop("c3type")
            self.devices[name] = dev_lib[dev_type](**props)
        self.chains = cfg["Chains"]
        self.__check_signal_chains()

    def write_config(self, filepath: str) -> None:
        """
        Write dictionary to a HJSON file.
        """
        with open(filepath, "w") as cfg_file:
            hjson.dump(self.asdict(), cfg_file, default=hjson_encode)

    def asdict(self) -> dict:
        """
        Return a dictionary compatible with config files.
        """
        devices = {}
        for name, dev in self.devices.items():
            devices[name] = dev.asdict()
        return {"Devices": devices, "Chains": self.chains}

    def __str__(self) -> str:
        return hjson.dumps(self.asdict(), default=hjson_encode)

    def generate_signals(self, instr: Instruction) -> dict:
        """
        Perform the signal chain for a specified instruction, including local
        oscillator, AWG generation and IQ mixing.

        Parameters
        ----------
        instr : Instruction
            Operation to be performed, e.g. logical gate.

        Returns
        -------
        dict
            Signal to be applied to the physical device.

        """
        gen_signal = {}
        gen_stacked_signals: dict = dict()
        for chan in instr.comps:
            signal_stack: List[tf.constant] = []
            gen_stacked_signals[chan] = []
            for dev_id in self.chains[chan]:
                dev = self.devices[dev_id]
                inputs = []
                for _input_num in range(dev.inputs):
                    inputs.append(signal_stack.pop())
                outputs = dev.process(instr, chan, *inputs)
                signal_stack.append(outputs)
                gen_stacked_signals[chan].append((dev_id, copy.deepcopy(outputs)))
            # The stack is reused here, thus we need to deepcopy.
            gen_signal[chan] = copy.deepcopy(signal_stack.pop())
        self.gen_stacked_signals = gen_stacked_signals

        # Hack to use crosstalk. Will be generalized to a post-processing module.
        # TODO: Rework of the signal generation for larger chips, similar to qiskit
        if "crosstalk" in self.devices:
            gen_signal = self.devices["crosstalk"].process(signal=gen_signal)
        return gen_signal



import time
import hjson
import tensorflow as tf
from c3.optimizers.optimalcontrol import OptimalControl
from c3.c3objs import hjson_encode
import numpy as np


class OptimalControlRobust(OptimalControl):
    """
    Object that deals with the open loop optimal control.

    Parameters
    ----------
    dir_path : str
        Filepath to save results
    fid_func : callable
        infidelity function to be minimized
    fid_subspace : list
        Indeces identifying the subspace to be compared
    pmap : ParameterMap
        Identifiers for the parameter vector
    callback_fids : list of callable
        Additional fidelity function to be evaluated and stored for reference
    algorithm : callable
        From the algorithm library
        Save plots of control signals
    store_unitaries : boolean
        Store propagators as text and pickle
    options : dict
        Options to be passed to the algorithm
    run_name : str
        User specified name for the run, will be used as root folder
    """

    def __init__(self, noise_map, **kwargs) -> None:
        super().__init__(**kwargs)
        self.noise_map = noise_map

    def goal_run_with_grad(self, current_params):
        goals = []
        goals_float = []
        grads = []
        evaluation = int(self.evaluation)
        for noise_vals, noise_map in self.noise_map:
            orig_val = np.array(self.exp.pmap.get_parameters([noise_map]))
            for noise_val in noise_vals:
                self.exp.pmap.set_parameters([noise_val], [noise_map])
                self.evaluation = evaluation
                with tf.GradientTape() as t:
                    t.watch(current_params)
                    goal = self.goal_run(current_params)
                grad = t.gradient(goal, current_params)
                goals.append(goal)
                goals_float.append(float(goal))
                grads.append(grad)
            self.exp.pmap.set_parameters(orig_val, [noise_map])

        self.optim_status["goals_individual"] = [float(goal) for goal in goals]
        self.optim_status["goal_std"] = float(tf.math.reduce_std(goals))
        self.optim_status["gradient_std"] = (
            tf.math.reduce_std(grads, axis=0).numpy().tolist()
        )
        self.optim_status["goal"] = float(tf.reduce_mean(goals, axis=0))
        self.optim_status["time"] = time.asctime()
        return tf.reduce_mean(goals, axis=0), tf.reduce_mean(grads, axis=0)

    def start_log(self):
        """
        Initialize the log with current time.

        """
        super().start_log()
        with open(self.logdir + self.logname, "a") as logfile:
            logfile.write("Robust values ")
            print(len(self.noise_map))
            logfile.write(hjson.dumps(self.noise_map, default=hjson_encode))
            logfile.write("\n")
            logfile.flush()



"""Integration Test module for Sensitivity Analysis on the dataset used for Model Learning
Please refer to test_model_learning.py for more details on the dataset
The test checks the integrity of the config file to ensure it creates the necessary optimizer
object which is then run. It also checks if the sweep goes till the higher end of the bounds, but
nothing else regarding the actual sweep run.
The integrity of the sweep algorithm is tested in test_scan_algos
"""

import hjson
import pytest
import numpy as np

from c3.optimizers.sensitivity import Sensitivity
from c3.experiment import Experiment

OPT_CONFIG_FILE_NAME = "test/sensitivity.cfg"
DESIRED_SWEEP_END_PARAMS = [-205000000.0, 5001500000.0]
SWEEP_PARAM_NAMES = ["Q1-anhar", "Q1-freq"]


@pytest.mark.integration
@pytest.mark.slow
def test_sensitivity() -> None:
    """Test sensitivity analysis with 1D sweeps on 2 variables"""
    with open(OPT_CONFIG_FILE_NAME, "r") as cfg_file:
        cfg = hjson.load(cfg_file)

    cfg.pop("optim_type")

    exp = Experiment()
    exp.read_config(cfg.pop("exp_cfg"))

    # test error handling for estimator
    with pytest.raises(NotImplementedError):
        opt = Sensitivity(**cfg, pmap=exp.pmap)

    cfg.pop("estimator")

    # test error handling for estimator_list
    with pytest.raises(NotImplementedError):
        opt = Sensitivity(**cfg, pmap=exp.pmap)

    cfg.pop("estimator_list")

    opt = Sensitivity(**cfg, pmap=exp.pmap)
    opt.set_exp(exp)
    opt.set_created_by(OPT_CONFIG_FILE_NAME)
    opt.run()

    for index, val in enumerate(SWEEP_PARAM_NAMES):
        # This decomposition of opt.sweep_end into the actual_param_end only makes
        # sense when you look at how opt.sweep_end structures the sweep endings
        actual_param_end = opt.sweep_end[index][val]["params"][0]
        np.testing.assert_almost_equal(
            actual_param_end, DESIRED_SWEEP_END_PARAMS[index]
        )
